@@ -13,34 +13,41 @@
 """Here is defined the Index class."""
 
 from __future__ import print_function
-import sys
-from bisect import bisect_left, bisect_right
-from time import time, clock
+from __future__ import absolute_import
+
+import math
+import operator
 import os
 import os.path
+import sys
 import tempfile
-import math
 import warnings
 
+from time import time, clock
+
 import numpy
 
-from tables.idxutils import (calc_chunksize, calcoptlevels,
+from .idxutils import (calc_chunksize, calcoptlevels,
                              get_reduction_level, nextafter, inftype)
 
-from tables import indexesextension
-from tables.node import NotLoggedMixin
-from tables.atom import UIntAtom, Atom
-from tables.earray import EArray
-from tables.carray import CArray
-from tables.leaf import Filters
-from tables.indexes import CacheArray, LastRowArray, IndexArray
-from tables.group import Group
-from tables.path import join_path
-from tables.exceptions import PerformanceWarning
-from tables.utils import is_idx, idx2long, lazyattr
-from tables.lrucacheextension import ObjectCache
+from . import indexesextension
+from .node import NotLoggedMixin
+from .atom import UIntAtom, Atom
+from .earray import EArray
+from .carray import CArray
+from .leaf import Filters
+from .indexes import CacheArray, LastRowArray, IndexArray
+from .group import Group
+from .path import join_path
+from .exceptions import PerformanceWarning
+from .utils import is_idx, idx2long, lazyattr
+from .utilsextension import (nan_aware_gt, nan_aware_ge,
+                                   nan_aware_lt, nan_aware_le,
+                                   bisect_left, bisect_right)
+from .lrucacheextension import ObjectCache
+import six
+from six.moves import range
 
-from tables._past import previous_api, previous_api_property
 
 
 # default version for INDEX objects
@@ -55,12 +62,13 @@ debug = False
 profile = False
 # profile = True  # Uncomment for profiling
 if profile:
-    from tables.utils import show_stats
+    from .utils import show_stats
 
 
 # The default method for sorting
-defsort = "quicksort"
-# defsort = "mergesort"
+# defsort = "quicksort"
+# Changing to mergesort to fix #441
+defsort = "mergesort"
 
 # Default policy for automatically updating indexes after a table
 # append operation, or automatically reindexing after an
@@ -97,10 +105,9 @@ def _table_column_pathname_of_index(indexpathname):
     colpathname = "/".join(names[i + 1:])
     return (tablepathname, colpathname)
 
-_tableColumnPathnameOfIndex = previous_api(_table_column_pathname_of_index)
 
-
-class Index(NotLoggedMixin, indexesextension.Index, Group):
+@six.python_2_unicode_compatible
+class Index(NotLoggedMixin, Group, indexesextension.Index):
     """Represents the index of a column in a table.
 
     This class is used to keep the indexing information for columns in a Table
@@ -154,27 +161,31 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
 
     _c_classid = 'INDEX'
 
-    _c_classId = previous_api_property('_c_classid')
 
-    # <properties>
-    kind = property(
-        lambda self: {1: 'ultralight', 2: 'light',
-                      4: 'medium', 8: 'full'}[self.indsize],
-        None, None,
-        "The kind of this index.")
+    @property
+    def kind(self):
+        "The kind of this index."
+        return {1: 'ultralight', 2: 'light',
+                4: 'medium', 8: 'full'}[self.indsize]
 
-    filters = property(
-        lambda self: self._v_filters, None, None,
+    @property
+    def filters(self):
         """Filter properties for this index - see Filters in
-        :ref:`FiltersClassDescr`.""")
+        :ref:`FiltersClassDescr`."""
+        return self._v_filters
 
-    def _getdirty(self):
-        if 'DIRTY' not in self._v_attrs:
-            # If there is no ``DIRTY`` attribute, index should be clean.
-            return False
-        return self._v_attrs.DIRTY
+    @property
+    def dirty(self):
+       """Whether the index is dirty or not.
+       Dirty indexes are out of sync with column data, so they exist but they
+       are not usable.
+       """
 
-    def _setdirty(self, dirty):
+       # If there is no ``DIRTY`` attribute, index should be clean.
+       return getattr(self._v_attrs, 'DIRTY', False)
+
+    @dirty.setter
+    def dirty(self, dirty):
         wasdirty, isdirty = self.dirty, bool(dirty)
         self._v_attrs.DIRTY = dirty
         # If an *actual* change in dirtiness happens,
@@ -185,47 +196,42 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
         if wasdirty and not isdirty:
             conditioncache.unnail()
 
-    dirty = property(
-        _getdirty, _setdirty, None,
-        """Whether the index is dirty or not.
+    @property
+    def column(self):
+        """The Column (see :ref:`ColumnClassDescr`) instance for the indexed
+        column."""
 
-        Dirty indexes are out of sync with column data, so they exist but they
-        are not usable.
-        """)
-
-    def _getcolumn(self):
         tablepath, columnpath = _table_column_pathname_of_index(
             self._v_pathname)
         table = self._v_file._get_node(tablepath)
         column = table.cols._g_col(columnpath)
         return column
 
-    column = property(_getcolumn, None, None,
-        """The Column (see :ref:`ColumnClassDescr`) instance for the indexed
-        column.""")
-
-    def _gettable(self):
-        tablepath, columnpath = _table_column_pathname_of_index(
-            self._v_pathname)
+    @property
+    def table(self):
+        """Accessor for the `Table` object of this index."""
+        tablepath, columnpath = _table_column_pathname_of_index(self._v_pathname)
         table = self._v_file._get_node(tablepath)
         return table
 
-    table = property(_gettable, None, None,
-                     "Accessor for the `Table` object of this index.")
+    @property
+    def nblockssuperblock(self):
+        "The number of blocks in a superblock."
+        return self.superblocksize // self.blocksize
 
-    nblockssuperblock = property(
-        lambda self: self.superblocksize // self.blocksize, None, None,
-        "The number of blocks in a superblock.")
+    @property
+    def nslicesblock(self):
+        "The number of slices in a block."
+        return self.blocksize // self.slicesize
 
-    nslicesblock = property(
-        lambda self: self.blocksize // self.slicesize, None, None,
-        "The number of slices in a block.")
+    @property
+    def nchunkslice(self):
+        "The number of chunks in a slice."
+        return self.slicesize // self.chunksize
 
-    nchunkslice = property(
-        lambda self: self.slicesize // self.chunksize, None, None,
-        "The number of chunks in a slice.")
-
-    def _g_nsuperblocks(self):
+    @property
+    def nsuperblocks(self):
+        "The total number of superblocks in index."
         # Last row should not be considered as a superblock
         nelements = self.nelements - self.nelementsILR
         nblocks = nelements // self.superblocksize
@@ -233,10 +239,9 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
             nblocks += 1
         return nblocks
 
-    nsuperblocks = property(_g_nsuperblocks, None, None,
-                            "The total number of superblocks in index.")
-
-    def _g_nblocks(self):
+    @property
+    def nblocks(self):
+        "The total number of blocks in index."
         # Last row should not be considered as a block
         nelements = self.nelements - self.nelementsILR
         nblocks = nelements // self.blocksize
@@ -244,34 +249,40 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
             nblocks += 1
         return nblocks
 
-    nblocks = property(_g_nblocks, None, None,
-                       "The total number of blocks in index.")
+    @property
+    def nslices(self):
+        "The number of complete slices in index."
+        return self.nelements // self.slicesize
 
-    nslices = property(
-        lambda self: self.nelements // self.slicesize, None, None,
-        "The number of complete slices in index.")
+    @property
+    def nchunks(self):
+        "The number of complete chunks in index."
+        return self.nelements // self.chunksize
 
-    nchunks = property(
-        lambda self: self.nelements // self.chunksize, None, None,
-        "The number of complete chunks in index.")
+    @property
+    def shape(self):
+        "The shape of this index (in slices and elements)."
+        return (self.nrows, self.slicesize)
 
-    shape = property(
-        lambda self: (self.nrows, self.slicesize), None, None,
-        "The shape of this index (in slices and elements).")
+    @property
+    def temp_required(self):
+        "Whether a temporary file for indexes is required or not."
+        return self.indsize > 1 and self.optlevel > 0 and self.table.nrows > self.slicesize
 
-    temp_required = property(
-        lambda self: (self.indsize > 1 and
-                      self.optlevel > 0 and
-                      self.table.nrows > self.slicesize),
-        None, None,
-        "Whether a temporary file for indexes is required or not.")
+    @property
+    def want_complete_sort(self):
+        "Whether we should try to build a completely sorted index or not."
+        return self.indsize == 8 and self.optlevel == 9
 
-    want_complete_sort = property(
-        lambda self: (self.indsize == 8 and self.optlevel == 9),
-        None, None,
-        "Whether we should try to build a completely sorted index or not.")
+    @property
+    def is_csi(self):
+        """Whether the index is completely sorted or not.
+
+        .. versionchanged:: 3.0
+           The *is_CSI* property has been renamed into *is_csi*.
+
+        """
 
-    def _is_csi(self):
         if self.nelements == 0:
             # An index with 0 indexed elements is not a CSI one (by definition)
             return False
@@ -286,18 +297,6 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
         self.compute_overlaps(self, None, False)
         return self.noverlaps == 0
 
-    _is_CSI = previous_api(_is_csi)
-
-    is_csi = property(_is_csi, None, None,
-        """Whether the index is completely sorted or not.
-
-        .. versionchanged:: 3.0
-           The *is_CSI* property has been renamed into *is_csi*.
-
-        """)
-
-    is_CSI = previous_api(is_csi)
-
     @lazyattr
     def nrowsinchunk(self):
         """The number of rows that fits in a *table* chunk."""
@@ -328,7 +327,6 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
             pass
         return lbucket
 
-    # </properties>
     def __init__(self, parentnode, name,
                  atom=None, title="",
                  kind=None,
@@ -395,7 +393,7 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
         self.tprof = 0
         """Time counter for benchmarking purposes."""
 
-        from tables.file import open_file
+        from .file import open_file
         self._openFile = open_file
         """The `open_file()` function, to avoid a circular import."""
 
@@ -555,7 +553,6 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
         if self.temp_required:
             self.create_temp()
 
-    _g_postInitHook = previous_api(_g_post_init_hook)
 
     def initial_append(self, xarr, nrow, reduction):
         """Compute an initial indices arrays for data to be indexed."""
@@ -596,7 +593,7 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
             idx = numpy.empty(len(arr), "uint%d" % (indsize * 8))
             lbucket = self.lbucket
             # Fill the idx with the bucket indices
-            offset = lbucket - ((nrow * (slicesize % lbucket)) % lbucket)
+            offset = int(lbucket - ((nrow * (slicesize % lbucket)) % lbucket))
             idx[0:offset] = 0
             for i in range(offset, slicesize, lbucket):
                 idx[i:i + lbucket] = (i + lbucket - 1) // lbucket
@@ -769,7 +766,6 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
         if profile:
             show_stats("Exiting appendLR", tref)
 
-    appendLastRow = previous_api(append_last_row)
 
     def optimize(self, verbose=False):
         """Optimize an index so as to allow faster searches.
@@ -1230,7 +1226,7 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
                 ul = self.nelementsILR // cs
                 bounds = numpy.concatenate((bounds, self.bebounds[:ul]))
             sbounds_idx = bounds.argsort(kind=defsort)
-            offset = nblock * nsb
+            offset = int(nblock * nsb)
             # Swap sorted and indices following the new order
             self.get_neworder(sbounds_idx, sorted, tmp_sorted, sortedLR,
                               nslices, offset, self.dtype)
@@ -1277,7 +1273,6 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
         stepl = numpy.array([1], dtype=numpy.uint64)
         where._g_write_slice(startl, stepl, countl, buffer)
 
-    read_sliceLR = previous_api(read_slice_lr)
 
     def reorder_slice(self, nslice, sorted, indices, ssorted, sindices,
                       tmp_sorted, tmp_indices):
@@ -1501,7 +1496,7 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
 
         # This method will only works under the assumtion that item
         # *is to be found* in the nslice.
-        assert limits[0] < item <= limits[1]
+        assert nan_aware_lt(limits[0], item) and nan_aware_le(item, limits[1])
         cs = self.chunksize
         ss = self.slicesize
         nelementsLR = self.nelementsILR
@@ -1737,7 +1732,6 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
 
         return self.read_sorted_indices('sorted', start, stop, step)
 
-    readSorted = previous_api(read_sorted)
 
     def read_indices(self, start=None, stop=None, step=None):
         """Return the indices values of index in the specified range.
@@ -1749,7 +1743,6 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
 
         return self.read_sorted_indices('indices', start, stop, step)
 
-    readIndices = previous_api(read_indices)
 
     def _process_range(self, start, stop, step):
         """Get a range specifc for the index usage."""
@@ -1771,7 +1764,6 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
             step = idx2long(step)
         return (start, stop, step)
 
-    _processRange = previous_api(_process_range)
 
     def __getitem__(self, key):
         """Return the indices values of index in the specified range.
@@ -1786,6 +1778,8 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
         """
 
         if is_idx(key):
+            key = operator.index(key)
+
             if key < 0:
                 # To support negative values
                 key += self.nelements
@@ -1954,8 +1948,8 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
 
         nchunk = -1
         # Lookup for item1
-        if item1 > b0:
-            if item1 <= b1:
+        if nan_aware_gt(item1, b0):
+            if nan_aware_le(item1, b1):
                 # Search the appropriate chunk in bounds cache
                 nchunk = bisect_left(bounds, item1)
                 # Lookup for this chunk in cache
@@ -1981,8 +1975,8 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
         else:
             start = 0
         # Lookup for item2
-        if item2 >= b0:
-            if item2 < b1:
+        if nan_aware_ge(item2, b0):
+            if nan_aware_lt(item2, b1):
                 # Search the appropriate chunk in bounds cache
                 nchunk2 = bisect_right(bounds, item2)
                 if nchunk2 != nchunk:
@@ -2011,7 +2005,6 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
             stop = 0
         return (start, stop)
 
-    searchLastRow = previous_api(search_last_row)
 
     def get_chunkmap(self):
         """Compute a map with the interesting chunks in index."""
@@ -2121,7 +2114,6 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
 
         return range_
 
-    getLookupRange = previous_api(get_lookup_range)
 
     def _f_remove(self, recursive=False):
         """Remove this Index object."""
@@ -2138,6 +2130,8 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
         if self.filters.complevel:
             if self.filters.shuffle:
                 filters += ", shuffle"
+            if self.filters.bitshuffle:
+                filters += ", bitshuffle"
             filters += ", %s(%s)" % (self.filters.complib,
                                      self.filters.complevel)
         return "Index(%s, %s%s).is_csi=%s" % \
@@ -2157,7 +2151,6 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
   slicesize := %s
   blocksize := %s
   superblocksize := %s
-  filters := %s
   dirty := %s
   byteorder := %r""" % (self._v_pathname, cpathname,
                         self.optlevel, self.kind,
@@ -2165,8 +2158,7 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
                         self.nelements,
                         self.chunksize, self.slicesize,
                         self.blocksize, self.superblocksize,
-                        self.filters, self.dirty,
-                        self.byteorder)
+                        self.dirty, self.byteorder)
         retstr += "\n  sorted := %s" % self.sorted
         retstr += "\n  indices := %s" % self.indices
         retstr += "\n  ranges := %s" % self.ranges
@@ -2179,7 +2171,6 @@ class Index(NotLoggedMixin, indexesextension.Index, Group):
 class IndexesDescG(NotLoggedMixin, Group):
     _c_classid = 'DINDEX'
 
-    _c_classId = previous_api_property('_c_classid')
 
     def _g_width_warning(self):
         warnings.warn(
@@ -2189,25 +2180,25 @@ class IndexesDescG(NotLoggedMixin, Group):
             "and possibly slow I/O" % self._v_max_group_width,
             PerformanceWarning)
 
-    _g_widthWarning = previous_api(_g_width_warning)
 
 
 class IndexesTableG(NotLoggedMixin, Group):
     _c_classid = 'TINDEX'
 
-    _c_classId = previous_api_property('_c_classid')
 
-    def _getauto(self):
+    @property
+    def auto(self):
         if 'AUTO_INDEX' not in self._v_attrs:
             return default_auto_index
         return self._v_attrs.AUTO_INDEX
 
-    def _setauto(self, auto):
+    @auto.setter
+    def auto(self, auto):
         self._v_attrs.AUTO_INDEX = bool(auto)
 
-    def _delauto(self):
+    @auto.deleter
+    def auto(self):
         del self._v_attrs.AUTO_INDEX
-    auto = property(_getauto, _setauto, _delauto)
 
     def _g_width_warning(self):
         warnings.warn(
@@ -2217,16 +2208,16 @@ class IndexesTableG(NotLoggedMixin, Group):
             "and possibly slow I/O" % self._v_max_group_width,
             PerformanceWarning)
 
-    _g_widthWarning = previous_api(_g_width_warning)
 
     def _g_check_name(self, name):
         if not name.startswith('_i_'):
             raise ValueError(
                 "names of index groups must start with ``_i_``: %s" % name)
 
-    _g_checkName = previous_api(_g_check_name)
 
-    def _gettable(self):
+    @property
+    def table(self):
+        "Accessor for the `Table` object of this `IndexesTableG` container."
         names = self._v_pathname.split("/")
         tablename = names.pop()[3:]   # "_i_" is at the beginning
         parentpathname = "/".join(names)
@@ -2234,17 +2225,12 @@ class IndexesTableG(NotLoggedMixin, Group):
         table = self._v_file._get_node(tablepathname)
         return table
 
-    table = property(
-        _gettable, None, None,
-        "Accessor for the `Table` object of this `IndexesTableG` container.")
-
 
 class OldIndex(NotLoggedMixin, Group):
     """This is meant to hide indexes of PyTables 1.x files."""
 
     _c_classid = 'CINDEX'
 
-    _c_classId = previous_api_property('_c_classid')
 
 
 ## Local Variables:
